from __future__ import annotations

from fmtr.tools import env, Path, net
from haco.paths import paths

STATUS = 'status'
HOSTNAME = net.get_hostname().lower()
MAC = net.MAC()

CLIENT_ID = f'{HOSTNAME}-{MAC.hex}'

TOPIC_ROOT = Path(paths.name) / env.CHANNEL
TOPIC_CLIENT = TOPIC_ROOT / CLIENT_ID
TOPIC_AVAIL = TOPIC_CLIENT / STATUS
PREFIX_MDI = "mdi:"



from dataclasses import dataclass

from haco.control import Control
from haco.utils import ConvertersString


@dataclass(kw_only=True)
class Select(Control):
    DATA = dict(
        platform='select'
    )
    converters = ConvertersString

    options: list[str]

    def command(self, value):
        raise NotImplementedError()

    def state(self, value):
        raise NotImplementedError()



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import ClassVar, TYPE_CHECKING, Callable

from fmtr.tools import aio
from haco.obs import logger

if TYPE_CHECKING:
    from capabilities import Capability


@dataclass(kw_only=True)
class AnnounceTopic:
    IO: ClassVar[str | None] = None

    KEY_DEFAULTS: ClassVar[dict] = {True: '{capability}_{io}_topic', False: '{io}_topic'}

    key: str | None = None
    value: str = '{path}/{capability}/{io}'

    parent: Capability | None = None
    announce: dict | None = field(default=None, metadata=dict(exclude=True))
    subscriptions: dict | None = field(default=None, metadata=dict(exclude=True), init=False)

    control_method: Callable | None = field(default=None, metadata=dict(exclude=True))

    def set_parent(self, capability: Capability):
        self.parent = capability
        if self.key is None:
            self.key = self.KEY_DEFAULTS[bool(self.capability.name)]

        method_name = self.callback_name
        method = getattr(self.capability.control, method_name, None)
        self.control_method = method
        setattr(self.capability.control, method_name, self.handle)

    def fill(self, mask):
        return mask.format(**self.fills)

    @property
    def fills(self):
        return dict(
            path=str(self.capability.control.topic),
            capability=self.capability.name or 'default',
            io=self.IO
        )

    @cached_property
    def topic_key(self):
        return self.fill(self.key)

    @cached_property
    def topic(self):
        return self.fill(self.value)

    def get_announce(self):
        return {self.topic_key: self.topic}

    @property
    def callback_name(self):
        if not self.capability.name:
            return self.IO

        return f'{self.capability.name}_{self.IO}'

    @property
    def callback_class_method_name(self):
        return f'{self.capability.control.__class__.__name__}.{self.callback_name}'

    def get_subscriptions(self):
        return {}

    @cached_property
    def capability(self):
        return self.parent

    async def handle(self, value):
        raise NotImplementedError()


@dataclass(kw_only=True)
class AnnounceTopicState(AnnounceTopic):
    IO: ClassVar[str] = 'state'

    def get_subscriptions(self):
        return {}

    async def handle(self, value=None):

        if not self.control_method:
            logger.error(f'Incomplete base class: {self.callback_class_method_name}')
            return

        is_async = aio.is_async(self.control_method)
        try:

            if is_async:
                value = await self.control_method(value)
            else:
                value = self.control_method(value)

        except NotImplementedError:
            logger.warning(f'Subclass has not implemented method: {self.callback_class_method_name} {self.topic} {value}')
            return

        value_raw = self.capability.converters.state(value)
        await self.capability.control.device.client.publish(self.topic, value_raw)
        return value_raw

@dataclass(kw_only=True)
class AnnounceTopicCommand(AnnounceTopic):
    IO: ClassVar[str] = 'command'

    @property
    def state(self):
        return self.capability.state

    async def handle(self, message):

        if not self.control_method:
            logger.error(f'Incomplete base class: {self.callback_class_method_name}')
            return

        is_async = aio.is_async(self.control_method)
        value = message.payload.decode('utf-8')
        value = self.capability.converters.command(value)

        try:

            if is_async:
                value_raw = await self.control_method(value)
            else:
                value_raw = self.control_method(value)

        except NotImplementedError:
            logger.warning(f'Subclass has not implemented method: {self.callback_class_method_name} {self.topic} {value}')
            return

        # Echo back as new state
        topic_state = self.state
        if topic_state:
            await topic_state.handle(value_raw)


    def get_subscriptions(self):
        return {self.topic: self}



from dataclasses import dataclass
from enum import StrEnum
from typing import ClassVar, Type

from haco.control import Control
from haco.uom import Uom
from haco.utils import Converters, ConvertersNumeric


class Mode(StrEnum):
    BOX = "box"
    SLIDER = "slider"


@dataclass(kw_only=True)
class Number(Control):
    DATA = dict(
        platform='number'
    )
    converters: ClassVar[Type[Converters]] = ConvertersNumeric

    min: float = 0.
    max: float = 100.
    mode: Mode = Mode.SLIDER
    step: float = 1.
    uom: Uom | None = None

    def command(self, value):
        raise NotImplementedError()

    def state(self, value):
        raise NotImplementedError()



from fmtr.tools import logging, debug, Constants
from haco.paths import paths
from haco.version import __version__

debug.trace()

logger = logging.get_logger(
    name=paths.name_ns,
    stream=Constants.INFRA,
    version=__version__,
)



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import List, TYPE_CHECKING

from haco.base import Base
from haco.control import Control
from haco.utils import sanitize_name

if TYPE_CHECKING:
    from haco.client import ClientHaco


@dataclass(kw_only=True)
class Device(Base):
    name: str
    manufacturer: str | None = None
    model: str | None = None
    sw_version: str | None = None

    identifiers: list[str] | None = field(default=None, init=False)

    controls: List[Control] = field(default_factory=list, metadata=dict(exclude=True))

    parent: ClientHaco | None = field(metadata=dict(exclude=True), init=False)
    subscriptions: dict | None = field(default=None, metadata=dict(exclude=True), init=False)


    def set_parent(self, client):
        self.parent = client

        self.identifiers = [self.name_san]

        for control in self.controls:
            control.set_parent(self)
        #self.announce = self.get_announce()
        self.subscriptions = self.get_subscriptions()

        self

    @cached_property
    def topic(self):
        return self.client.topic / self.name_san

    @property
    def name_san(self) -> str:
        return sanitize_name(self.name)

    async def announce(self):
        for control in self.controls:
            await control.announce()

    async def initialise(self):
        for control in self.controls:
            await control.initialise()


    @property
    def client(self):
        return self.parent

    def get_subscriptions(self) -> dict:
        data = {}
        for control in self.controls:
            data |= control.subscriptions

        return data


if __name__ == "__main__":
    control = Control(name="test", platform="test")
    device = Device(name="test", parent=None, controls=[control])
    data = device.model_dump()
    print(device.topic)



import asyncio
import logging

from haco.button import Button
from haco.client import ClientHaco
from haco.climate import Climate
from haco.constants import MQTT_HOST
from haco.device import Device
from haco.obs import logger
from haco.select import Select
from haco.switch import Switch

handler = logger.LogfireLoggingHandler()
logging.basicConfig(handlers=[handler], level=logging.DEBUG)

async def main():
    climate_a = Climate(name="Dev Climate A", fan_modes=['a', 'b', 'c'])
    climate_b = Climate(name="Dev Climate B", fan_modes=['a', 'b', 'c'])

    b1 = Button(name="Dev Button 1")

    s1 = Switch(name="Dev Switch 1")
    sel1 = Select(name="Dev Select 1", options=['a', 'b', 'c'])
    device = Device(name="Dev Device", controls=[climate_a, climate_b, b1, s1, sel1])

    client = ClientHaco(hostname=MQTT_HOST, device=device, logger=logging.getLogger(__name__))

    await client.start()




if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypeVar, Generic, Type, ClassVar

from fmtr.tools.json_tools import to_json
from haco.base import Base
from haco.capabilities import Capability
from haco.constants import PREFIX_MDI
from haco.obs import logger
from haco.utils import sanitize_name, Converters, ConvertersBool

if TYPE_CHECKING:
    from haco.device import Device

DeviceT = TypeVar("DeviceT", bound="Device")


@dataclass(kw_only=True)
class Control(Base, Generic[DeviceT]):
    converters: ClassVar[Type[Converters]] = ConvertersBool

    name: str
    icon: str | None = None
    device: DeviceT = field(default=None, init=False)
    capabilities: None | list[Capability] = field(default=None, metadata=dict(exclude=True))
    unique_id: str | None = field(default=None, init=False)
    availability_topic: str | None = field(default=None, init=False)

    # announce: dict | None = field(default=None, metadata=dict(exclude=True))
    subscriptions: dict | None = field(default=None, metadata=dict(exclude=True), init=False)

    def __post_init__(self):
        self.capabilities = self.get_capabilities()

        if not self.icon:
            return

        if not self.icon.startswith(PREFIX_MDI):
            self.icon = f"{PREFIX_MDI}{self.icon}"


    def set_parent(self, device):
        self.device = device

        for cap in self.capabilities:
            cap.set_parent(self)

        self.unique_id = self.get_unique_id()
        self.availability_topic = self.get_availability_topic()
        self.subscriptions = self.get_subscriptions()
        #self.announce = self.get_announce()



    @property
    def parent(self):
        return self.device

    def get_announce(self):
        data = self.model_dump()
        for cap in self.capabilities:
            data |= cap.get_announce()

        return data

    @classmethod
    def get_capabilities(cls):
        return [
            Capability(name=None)
        ]

    def get_unique_id(self) -> str:
        return f"{self.device.name_san}-{self.name_san}"

    @property
    def name_san(self) -> str:
        return sanitize_name(self.name)

    @property
    def topic(self):
        return self.device.topic / self.name_san

    @property
    def announce_topic(self):
        return f"homeassistant/{self.DATA['platform']}/{self.unique_id}/config"

    def get_availability_topic(self) -> str:
        return self.device.client.will.topic

    def get_subscriptions(self) -> dict:

        data = {}

        for capability in self.capabilities:
            data |= capability.get_subscriptions()

        return data

    async def initialise(self):
        await self.announce()
        for capability in self.capabilities:
            if not capability.state:
                continue
            await capability.state.handle(value=None)

    async def announce(self):
        """

        Entities must be able to announce themselves to Home Assistant, e.g. when a pulldown changes its options, etc.

        """
        topic = self.announce_topic
        data_json = to_json(self.get_announce())
        logger.info(f'Announcing {topic} with {data_json}')
        await self.device.client.publish(topic, data_json, retain=True)



if __name__ == "__main__":
    control = Control(name="test", platform="test")
    control



from fmtr.tools import Setup

setup = Setup(
    org=None,
    dependencies=dict(
        install=[
            'fmtr.tools[version.dev,logging,sets,yaml,debug,caching,api,mqtt]',
        ]
    ),
    description='Home Assistant Control Objects',
    keywords='homeassistant controls python tasmota',
    # url=f'https://fmtr.link/{name}',
)
